//
//  EventViewController.swift
//  Teleasystent
//

import UIKit
import PopupDialog

class EventViewController: UIViewController {
    @IBOutlet weak var assistantLabel: UILabel!
    @IBOutlet weak var dateLabel: UILabel!
    @IBOutlet weak var typeLabel: UILabel!
    @IBOutlet weak var callButton: UIButton!
    @IBOutlet weak var deleteButton: UIButton!
    @IBOutlet weak var editButton: UIButton!
    @IBOutlet weak var markButton: UIButton!
    
    var event: EventData?
    
    override func viewDidLoad() {
        super.viewDidLoad()

        // Do any additional setup after loading the view.
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        guard let event = event else {
            navigationController?.popViewController(animated: false)
            return
        }
        
        UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, assistantLabel)
        
        self.title = Utils.getLocString("event_title")
        assistantLabel.text = EventsListViewController.assistantName(event: event)
        dateLabel.text = event.visibleDate
        typeLabel.text = AppDelegate.getInstance().serverManager.eventTypes?[event.typeId]?.locName
        
        callButton.setTitle(Utils.getLocString("event_call"), for: .normal)
        deleteButton.setTitle(Utils.getLocString("event_delete"), for: .normal)
        editButton.setTitle(Utils.getLocString("event_edit"), for: .normal)
        markButton.setTitle(Utils.getLocString("event_mark"), for: .normal)
        
        dateLabel.isAccessibilityElement = false
        typeLabel.isAccessibilityElement = false
        assistantLabel.accessibilityLabel = Utils.getLocString("event_label")
        assistantLabel.accessibilityValue = "\(assistantLabel.text ?? ""),,\(event.visibleDateHint),,\(typeLabel.text ?? "")"
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        
        assistantLabel.accessibilityFrame = assistantLabel.accessibilityFrame.union(dateLabel.accessibilityFrame.union(typeLabel.accessibilityFrame))
        
        restoreEventIfPossible()
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        if isBeingDismissed || isMovingFromParentViewController {
            AppDelegate.getInstance().restoreData.selectedEventId = nil
        }
    }
    
    func restoreEventIfPossible() {
        if AppDelegate.getInstance().restoreData.addEditLaunched {
            editButton.sendActions(for: .touchUpInside)
        }
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    @IBAction func onCallButtonClicked(_ sender: UIButton) {
        guard let assistant = AppDelegate.getInstance().serverManager.assistants?[event?.assistantId ?? ""] else {
            let _ = presentErrorDialog(title: nil, message: Utils.getLocString("event_callAssistantError"), okButton: Utils.getLocString("common_ok") ?? "ok") {
                weak var weakSelf = self
                weakSelf?.navigationController?.popToRootViewController(animated: true)
                AppDelegate.getInstance().restoreData.clear()
            }
            return
        }
        if assistant.assistantId == "0" {
            let _ = presentErrorDialog(title: nil, message: Utils.getLocString("event_callAssistantNotAssignedError"), okButton: Utils.getLocString("common_ok") ?? "ok") { }
            return
        }
        
        let skypeOpened = UIApplication.shared.openURL(URL(string: "skype:\(assistant.skypeName)?call&video=true")!)
        if !skypeOpened {
            UIApplication.shared.openURL(URL(string: Utils.appITunesLink("304878510", useHttp: false))!)
        } else {
            AppDelegate.getInstance().locationManager.start()
        }
    }
    
    @IBAction func onDeleteButtonClicked(_ sender: UIButton) {
        let alert = UIAlertController(title: nil, message: Utils.getLocString("event_deleteConfirmMessage"), preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: Utils.getLocString("common_yes"), style: .default, handler: { action in
            weak var weakSelf = self
            weakSelf?.deleteEvent()
        }))
        alert.addAction(UIAlertAction(title: Utils.getLocString("common_cancel"), style: .cancel, handler: nil))
        present(alert, animated: true, completion: nil)
    }
    
    func deleteEvent() {
        let loadingVC = LoadingDialogController()
        present(loadingVC, animated: false, completion: nil)
        AppDelegate.getInstance().serverManager.deleteEvent(id: event!.eventId!) { success, error in
            weak var weakSelf = self
            Async.main {
                loadingVC.dismiss(animated: false, completion: nil)
                if success {
                    weakSelf?.onDeleteEventCompleted()
                } else {
                    let error = error ?? NSError(domain: "DeleteEvent", code: 1, userInfo: nil)
                    weakSelf?.handleDeleteEventError(error: error)
                }
            }
        }
    }
    
    func setEventsListRefreshOnAppear(refresh: Bool = true) {
        if let viewControllers = navigationController?.viewControllers {
            for vc in viewControllers {
                if let eventsListVC = vc as? EventsListViewController {
                    eventsListVC.refreshListOnAppear = refresh
                }
            }
        }
    }
    
    func onDeleteEventCompleted() {
        setEventsListRefreshOnAppear()
        navigationController?.popViewController(animated: true)
    }
    
    func handleDeleteEventError(error: Error) {
        let _ = presentErrorDialog(error: error) { }
    }
    
    @IBAction func onEditButtonClicked(_ sender: UIButton) {
        AppDelegate.getInstance().restoreData.addEditLaunched = true
    }
    
    @IBAction func onMarkButtonClicked(_ sender: UIButton) {
        let ratingVC = RatingViewController(nibName: "RatingViewController", bundle: nil)
        if let mark = event?.markInt, mark > 0 {
            ratingVC.defaultRating = mark
        }
        let popup = PopupDialog(viewController: ratingVC,
                                buttonAlignment: .horizontal,
                                transitionStyle: .bounceDown,
                                tapGestureDismissal: true,
                                panGestureDismissal: false)
        
        let buttonOne = CancelButton(title: Utils.getLocString("common_cancel")!, height: 60) {
            print("mark canceled")
        }
        let buttonTwo = DefaultButton(title: Utils.getLocString("rate_rate")!, height: 60) {
            let rating = ratingVC.cosmosStarRating.rating
            weak var weakSelf = self
            popup.dismiss(animated: false, completion: nil)
            weakSelf?.onMarkConfirmed(mark: Int(rating))
        }
        popup.addButtons([buttonOne, buttonTwo])
        present(popup, animated: true, completion: nil)
    }
    
    func onMarkConfirmed(mark: Int) {
        let loadingVC = LoadingDialogController()
        present(loadingVC, animated: false, completion: nil)
        event!.mark = String(mark)
        AppDelegate.getInstance().serverManager.addEditEvent(event: event!) { eventData, eventId, error in
            weak var weakSelf = self
            Async.main {
                loadingVC.dismiss(animated: false, completion: nil)
                weakSelf?.setEventsListRefreshOnAppear()
                if let error = error {
                    weakSelf?.handleMarkError(error: error)
                }
            }
        }
    }
    
    func handleMarkError(error: Error) {
        let _ = presentErrorDialog(error: error) { }
    }
    
    
    // MARK: - Navigation

    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let addEditEventVC = segue.destination as? AddEditEventViewController {
            addEditEventVC.event = event
        }
    }
    

}
