//
//  EventsListViewController.swift
//  Teleasystent
//

import UIKit

class EventsListViewController: UIViewController, UITableViewDataSource, UITableViewDelegate {
    
    @IBOutlet weak var addEventButton: UIBaseButton!
    @IBOutlet weak var eventsLabel: BaseUILabel!
    @IBOutlet weak var eventsTableView: UITableView!
    
    var user: UserData?
    var refreshListOnAppear = true
    
    static let noAssistantName = Utils.getLocString("addevent_assistant_none")!
    static let noEventTypeName = Utils.getLocString("addevent_type_none")!
    let headerBgColor = UIColor(white: 0.85, alpha: 1)
    let cellBgColor = UIColor(white: 1.0, alpha: 1)
    let cellPressedBgColor = UIColor(white: 0.8, alpha: 1)
    
    private var events = [EventData]()
    private var lastEventsUpdateTime: Date?
    private var lastSelectedEvent: EventData?
    private var sortColumn = SortColumn.Date
    private var sortAscending = false
    private var headerCell : EventsListTableViewCell?
    private weak var refreshEventsErrorDialog: UIAlertController?
    
    enum SortColumn {
        case Assistant
        case Date
        case EventType
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        //eventsTableView.tableFooterView = UIView()
        eventsTableView.dataSource = self
        eventsTableView.delegate = self
    }
    
    override func viewWillAppear(_ animated: Bool) {
        print("will appear")
        super.viewWillAppear(animated)
        
        navigationItem.leftBarButtonItem = UIBarButtonItem(title: Utils.getLocString("eventslist_signout")!, style: .plain, target: self, action: #selector(onSignOutPressed))
        navigationItem.rightBarButtonItem = UIBarButtonItem(barButtonSystemItem: UIBarButtonSystemItem.refresh, target: self, action: #selector(onRefreshPressed))
        navigationItem.rightBarButtonItem?.accessibilityLabel = Utils.getLocString("eventslist_refresh")
        
        self.title = Utils.getLocString("eventslist_title")
        addEventButton.setTitle(Utils.getLocString("eventslist_add"), for: .normal)
        eventsLabel.text = Utils.getLocString("eventslist_listLabel")
        
        UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, addEventButton)
        
        let _ = AppDelegate.getInstance().serverManager.updateAssistantsIfStale(completion: nil)
        let _ = AppDelegate.getInstance().serverManager.updateEventTypesIfStale(completion: nil)
        if refreshListOnAppear || AppDelegate.getInstance().shouldUpdateDataFromServer(lastUpdate: lastEventsUpdateTime) {
            refreshEvents()
        }
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    func restoreEventIfPossible() {
        if let eventId = AppDelegate.getInstance().restoreData.selectedEventId {
            guard let event = events.first(where: { $0.eventId?.compare(eventId) == .orderedSame }) else {
                return
            }
            onEventSelected(event: event)
        } else if AppDelegate.getInstance().restoreData.addEditLaunched {
            addEventButton.sendActions(for: .touchUpInside)
        }
    }
    
    func refreshEvents() {
        if let errorDialog = refreshEventsErrorDialog {
            errorDialog.dismiss(animated: false, completion: nil)
            refreshEventsErrorDialog = nil
        }
        
        let loadingVC = LoadingDialogController()
        present(loadingVC, animated: false, completion: nil)
        AppDelegate.getInstance().serverManager.getAllEvents(assistantId: nil, userId: user!.userId) { events, error in
            weak var weakSelf = self
            Async.main {
                loadingVC.dismiss(animated: false, completion: nil)
                UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, weakSelf?.addEventButton)
                if let error = error {
                    weakSelf?.handleGetEventsError(error: error)
                } else if let events = events {
                    weakSelf?.lastEventsUpdateTime = Date(timeIntervalSinceNow: 0.0)
                    weakSelf?.reloadEventsList(events: events)
                    weakSelf?.restoreEventIfPossible()
                } else {
                    weakSelf?.handleGetEventsError(error: NSError(domain: "GetEventError", code: 1, userInfo: nil))
                }
                weakSelf?.refreshListOnAppear = false
            }
        }
    }
    
    func reloadEventsList(events: [EventData]?) {
        if let events = events {
            switch sortColumn {
            case .Assistant: self.events = events.sorted(by: {
                let left = EventsListViewController.assistantName(event: $0)
                let right = EventsListViewController.assistantName(event: $1)
                return sortAscending ? left < right : left > right
            })
            case .Date: self.events = events.sorted(by: { sortAscending ? $0.date < $1.date : $0.date > $1.date })
            case .EventType: self.events = events.sorted(by: {
                let left = EventsListViewController.eventTypeName(event: $0)
                let right = EventsListViewController.eventTypeName(event: $1)
                return sortAscending ? left < right : left > right
            })
            }
        }
        eventsTableView.reloadData()
        updateHeaderUI()
    }
    
    func handleGetEventsError(error: Error) {
        print("GetEvents error: \(error)")
        refreshEventsErrorDialog = presentErrorDialog(error: error) {
            weak var weakSelf = self
            weakSelf?.refreshEvents()
        }
    }
    
    @objc func onSignOutPressed() {
        AppDelegate.getInstance().firebaseAuthManager.logout()
        navigationController?.popViewController(animated: true)
    }
    
    @objc func onRefreshPressed() {
        let _ = AppDelegate.getInstance().serverManager.updateAssistantsIfStale(completion: nil)
        let _ = AppDelegate.getInstance().serverManager.updateEventTypesIfStale(completion: nil)
        refreshEvents()
    }
    
    @IBAction func onAddEventButtonClicked(_ sender: UIButton) {
        AppDelegate.getInstance().restoreData.addEditLaunched = true
    }
    
    func createTableHeader() -> EventsListTableViewCell {
        let cell = eventsTableView.dequeueReusableCell(withIdentifier: "EventsListCell") as! EventsListTableViewCell
        headerCell = cell
        cell.backgroundColor = headerBgColor
        
        let heightMult: CGFloat = 0.4
        cell.assistant.fitToHeightMult = heightMult
        cell.date.fitToHeightMult = heightMult
        cell.type.fitToHeightMult = heightMult
        
        cell.assistant.isUserInteractionEnabled = true
        cell.assistant.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(onAssistantHeaderClicked)))
        
        cell.date.isUserInteractionEnabled = true
        cell.date.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(onDateHeaderClicked)))
        
        cell.type.isUserInteractionEnabled = true
        cell.type.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(onEventTypeHeaderClicked)))
        
        cell.separatorInset = UIEdgeInsets.zero
        cell.layoutMargins = UIEdgeInsets.zero
        
        updateHeaderUI()
        
        cell.isAccessibilityElement = false
        for label in [cell.assistant, cell.date, cell.type] as [UILabel] {
            label.isAccessibilityElement = true
            label.accessibilityTraits = UIAccessibilityTraitHeader
            label.accessibilityHint = Utils.getLocString("eventslist_sortHint")
        }
        
        return cell
    }
    
    @objc func onAssistantHeaderClicked() {
        onHeaderClicked(column: .Assistant)
    }
    
    @objc func onDateHeaderClicked() {
        onHeaderClicked(column: .Date)
    }
    
    @objc func onEventTypeHeaderClicked() {
        onHeaderClicked(column: .EventType)
    }
    
    func onHeaderClicked(column: SortColumn) {
        let oldSortColumn = sortColumn
        sortColumn = column
        if oldSortColumn == sortColumn {
            sortAscending = !sortAscending
        } else {
            sortAscending = true
        }
        reloadEventsList(events: events)
    }
    
    func updateHeaderUI() {
        if let cell = headerCell {
            cell.assistant.text = Utils.getLocString("eventslist_assistantLabel")
            cell.date.text = Utils.getLocString("eventslist_dateLabel")
            cell.type.text = Utils.getLocString("eventslist_typeLabel")
            
            for label in [cell.assistant, cell.date, cell.type] as [UILabel] {
                label.accessibilityLabel = label.text
                label.accessibilityValue = nil
            }
            
            let sortedPrefix = sortAscending ? "\u{25B2} " : "\u{25BC} "
            let sortedValue = Utils.getLocString(sortAscending ? "eventslist_sortAscValue" : "eventslist_sortDescValue")
            switch sortColumn {
            case .Assistant:
                cell.assistant.text = "\(sortedPrefix)\(cell.assistant.text ?? "")"
                cell.assistant.accessibilityValue = sortedValue
            case .Date:
                cell.date.text = "\(sortedPrefix)\(cell.date.text ?? "")"
                cell.date.accessibilityValue = sortedValue
            case .EventType:
                cell.type.text = "\(sortedPrefix)\(cell.type.text ?? "")"
                cell.type.accessibilityValue = sortedValue
            }
        }
    }
    
    
    func numberOfSections(in tableView: UITableView) -> Int {
        return 1
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return events.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "EventsListCell", for: indexPath) as! EventsListTableViewCell
        let event = events[indexPath.row]
        
        cell.backgroundColor = cellBgColor
        cell.assistant.text = EventsListViewController.assistantName(event: event)
        cell.date.text = event.visibleDate
        cell.type.text = EventsListViewController.eventTypeName(event: event)
        
        cell.date.accessibilityLabel = event.visibleDateHint
        
        cell.accessibilityHint = Utils.getLocString("eventslist_selectHint")
        
        cell.separatorInset = UIEdgeInsets.zero
        cell.layoutMargins = UIEdgeInsets.zero
        return cell
    }
    
    static func assistantName(event: EventData) -> String {
        return assistantName(assistantId: event.assistantId ?? "0")
    }
    
    static func assistantName(assistantId: String) -> String {
        var name = noAssistantName
        if assistantId != "0" {
            name = AppDelegate.getInstance().serverManager.assistants?[assistantId]?.name ?? name
        }
        return name
    }
    
    static func eventTypeName(event: EventData) -> String {
        return AppDelegate.getInstance().serverManager.eventTypes?[event.typeId]?.locName ?? noEventTypeName
    }
    
    func tableView(_ tableView: UITableView, viewForHeaderInSection section: Int) -> UIView? {
        return createTableHeader()
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return UIScreen.main.bounds.height / 9
    }
    
    func tableView(_ tableView: UITableView, heightForHeaderInSection section: Int) -> CGFloat {
        return UIScreen.main.bounds.height / 15
    }
    
    func tableView(_ tableView: UITableView, didHighlightRowAt indexPath: IndexPath) {
        let cell = tableView.cellForRow(at: indexPath)
        cell?.backgroundColor = cellPressedBgColor
    }
    
    func tableView(_ tableView: UITableView, didUnhighlightRowAt indexPath: IndexPath) {
        let cell = tableView.cellForRow(at: indexPath)
        cell?.backgroundColor = cellBgColor
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        onEventSelected(event: events[indexPath.row])
    }
    
    func onEventSelected(event: EventData) {
        lastSelectedEvent = event
        performSegue(withIdentifier: "showEventSegue", sender: self)
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let eventVC = segue.destination as? EventViewController {
            eventVC.event = lastSelectedEvent
            AppDelegate.getInstance().restoreData.selectedEventId = lastSelectedEvent?.eventId
        }
    }



}
