//
//  LocationManager.swift
//  Teleasystent
//

import UIKit
import CoreLocation

class LocationManager: NSObject, CLLocationManagerDelegate {
    
    private let manager: CLLocationManager
    private var started = false
    private var lastSendTimestamp: Date?
    private var stopTime: Date?
    
    override init() {
        manager = CLLocationManager()
        super.init()
        
        manager.delegate = self
        manager.desiredAccuracy = kCLLocationAccuracyBest
        manager.allowsBackgroundLocationUpdates = true
    }
    
    func getAuthorizationStatus() -> CLAuthorizationStatus {
        return CLLocationManager.authorizationStatus()
    }
    
    func isLocationServiceEnabled() -> Bool {
        return CLLocationManager.locationServicesEnabled()
    }
    
    func requestAuthorizationIfNecessary() {
        switch getAuthorizationStatus() {
        case .notDetermined, .authorizedWhenInUse: manager.requestAlwaysAuthorization()
        case .restricted, .denied: onLocationError(error: "permission not granted")
        case .authorizedAlways: break
        }
    }
    
    private func onLocationError(error: String) {
        stop()
        if let userId = AppDelegate.getInstance().firebaseAuthManager.currentUser?.uid {
            let locationData = LocationData(userId: userId, error: error, curPosTimestamp: Date(timeIntervalSinceNow: 0.0))
            AppDelegate.getInstance().serverManager.editLocation(location: locationData, completion: nil)
        }
    }
    
    func start(runTimeSeconds: Double = 120.0) {
        if !started {
            requestAuthorizationIfNecessary()
            manager.startUpdatingLocation()
            started = true
        }
        stopTime = Date(timeIntervalSinceNow: runTimeSeconds)
    }
    
    func stop() {
        if started {
            manager.stopUpdatingLocation()
            started = false
        }
        stopTime = nil
    }
    
    internal func locationManager(_ manager: CLLocationManager, didUpdateLocations locations: [CLLocation]) {
        if let stopTime = stopTime, stopTime.timeIntervalSinceNow <= 0 {
            stop()
        } else {
            let location = locations.last!
            if shouldSendLocation(), let userId = AppDelegate.getInstance().firebaseAuthManager.currentUser?.uid {
                let locationData = LocationData(userId: userId, curPosLatitude: location.coordinate.latitude, curPosLongitude: location.coordinate.longitude, curPosTimestamp: location.timestamp)
                AppDelegate.getInstance().serverManager.editLocation(location: locationData, completion: nil)
                lastSendTimestamp = Date(timeIntervalSinceNow: 0.0)
            }
//            print("Location: timestamp=\(location.timestamp), latitude=\(location.coordinate.latitude), longitude=\(location.coordinate.longitude)")
        }
    }
    
    internal func locationManager(_ manager: CLLocationManager, didChangeAuthorization status: CLAuthorizationStatus) {
        requestAuthorizationIfNecessary()
    }
    
    internal func locationManager(_ manager: CLLocationManager, didFailWithError error: Error) {
        requestAuthorizationIfNecessary()
        if !isLocationServiceEnabled() {
            onLocationError(error: "location service disabled")
        }
    }
    
    private func shouldSendLocation() -> Bool {
        if let lastUpdate = lastSendTimestamp {
            return lastUpdate.timeIntervalSinceNow < -5
        }
        return true
    }
}
