//
//  EventViewController.swift
//  Teleasystent
//

import UIKit
import PopupDialog

class EventViewController: UIViewController {
    
    @IBOutlet weak var userLabel: BaseUILabel!
    @IBOutlet weak var dateLabel: UILabel!
    @IBOutlet weak var typeLabel: UILabel!
    @IBOutlet weak var stickLabel: BaseUILabel!
    @IBOutlet weak var dogImage: UIImageView!
    @IBOutlet weak var descriptionText: UITextView!
    @IBOutlet weak var callButton: UIButton!
    @IBOutlet weak var assignButton: UIBaseButton!
    @IBOutlet weak var mapButton: UIBaseButton!
    
    var event: EventData?
    var eventUser: UserData?
    
    private var assistantId: String!
    
    override func viewDidLoad() {
        super.viewDidLoad()

        descriptionText.layer.borderColor = UIColor(white: 204.0/255.0, alpha: 1.0).cgColor
        descriptionText.layer.borderWidth = 0.5
        descriptionText.layer.cornerRadius = 5.0
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        guard let event = event, let eventUser = eventUser, let assistantId = AppDelegate.getInstance().firebaseAuthManager.currentUser?.uid else {
            navigationController?.popViewController(animated: false)
            return
        }
        self.assistantId = assistantId
        
        UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, userLabel)
        
        self.title = Utils.getLocString("event_title")
        userLabel.text = eventUser.name
        dateLabel.text = event.visibleDate
        typeLabel.text = AppDelegate.getInstance().serverManager.eventTypes?[event.typeId]?.locName
        stickLabel.text = Utils.getLocString(eventUser.stickInRightHand ? "event_stickRight" : "event_stickLeft")
        dogImage.isHidden = !eventUser.hasDog
        var description = event.description ?? ""
        if description.trimmingCharacters(in: .whitespaces).isEmpty {
            description = Utils.getLocString("event_descriptionNone")!
        }
        descriptionText.text = "\(Utils.getLocString("addevent_descriptionLabel")!): \(description)"
        
        callButton.setTitle(Utils.getLocString("event_call"), for: .normal)
        updateAssignUI()
        mapButton.setTitle(Utils.getLocString("event_location"), for: .normal)
        
        dateLabel.isAccessibilityElement = false
        typeLabel.isAccessibilityElement = false
        userLabel.accessibilityLabel = Utils.getLocString("event_label")
        userLabel.accessibilityValue = "\(userLabel.text ?? ""),,\(event.visibleDateHint),,\(typeLabel.text ?? "")"
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        
        userLabel.accessibilityFrame = userLabel.accessibilityFrame.union(dateLabel.accessibilityFrame.union(typeLabel.accessibilityFrame))
        
        restoreEventIfPossible()
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        if isBeingDismissed || isMovingFromParentViewController {
            AppDelegate.getInstance().restoreData.selectedEventId = nil
        }
    }
    
    func restoreEventIfPossible() {
        if AppDelegate.getInstance().restoreData.mapLaunched {
            mapButton.sendActions(for: .touchUpInside)
        }
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    func isEventAssignedToSelf() -> Bool {
        guard let event = event else {
            return false
        }
        return event.assistantId == assistantId
    }
    
    func updateAssignUI() {
        assignButton.setTitle(Utils.getLocString(isEventAssignedToSelf() ? "event_unassign" : "event_assign"), for: .normal)
    }
    
    @IBAction func onCallButtonClicked(_ sender: UIButton) {
        guard let eventUser = eventUser else {
            let _ = presentErrorDialog(title: nil, message: Utils.getLocString("event_callUserError"), okButton: Utils.getLocString("common_ok") ?? "ok") {
                weak var weakSelf = self
                weakSelf?.navigationController?.popToRootViewController(animated: true)
                AppDelegate.getInstance().restoreData.clear()
            }
            return
        }
        
        let skypeOpened = UIApplication.shared.openURL(URL(string: "skype:\(eventUser.skypeName)?call&video=true")!)
        if !skypeOpened {
            UIApplication.shared.openURL(URL(string: Utils.appITunesLink("304878510", useHttp: false))!)
        } else {
            LocationViewController.sendLocationNotification(userId: eventUser.userId)
        }
    }
    
    @IBAction func onAssignButtonClicked(_ sender: UIButton) {
        if isEventAssignedToSelf() {
            let alert = UIAlertController(title: nil, message: Utils.getLocString("event_unassignConfirmMessage"), preferredStyle: .alert)
            alert.addAction(UIAlertAction(title: Utils.getLocString("common_yes"), style: .default, handler: { action in
                weak var weakSelf = self
                weakSelf?.assignAssistantToEvent(assistantId: AssistantData.assistantNotAssignedId)
            }))
            alert.addAction(UIAlertAction(title: Utils.getLocString("common_cancel"), style: .cancel, handler: nil))
            present(alert, animated: true, completion: nil)
        } else {
            assignAssistantToEvent(assistantId: assistantId)
        }
    }
    
    func assignAssistantToEvent(assistantId: String) {
        let loadingVC = LoadingDialogController()
        present(loadingVC, animated: false, completion: nil)
        event!.assistantId = assistantId
        AppDelegate.getInstance().serverManager.addEditEvent(event: event!) { eventData, eventId, error in
            weak var weakSelf = self
            Async.main {
                loadingVC.dismiss(animated: false, completion: nil)
                if let error = error {
                    weakSelf?.handleAssignAssistantToEventError(error: error)
                } else if let evId = eventId {
                    weakSelf?.onAssignAssistantToEventCompleted(event: eventData, eventId: evId)
                } else {
                    weakSelf?.handleAssignAssistantToEventError(error: NSError(domain: "AssignToEvent", code: 1, userInfo: nil))
                }
            }
        }
    }
    
    func handleAssignAssistantToEventError(error: Error) {
        let _ = presentErrorDialog(error: error) { }
    }
    
    func onAssignAssistantToEventCompleted(event: EventData, eventId: String) {
        updateAssignUI()
        setEventsListRefreshOnAppear()
    }
    
    @IBAction func onMapClicked(_ sender: UIBaseButton) {
        AppDelegate.getInstance().restoreData.mapLaunched = true
    }
    
    
    func setEventsListRefreshOnAppear(refresh: Bool = true) {
        if let viewControllers = navigationController?.viewControllers {
            for vc in viewControllers {
                if let eventsListVC = vc as? EventsListViewController {
                    eventsListVC.refreshListOnAppear = refresh
                }
            }
        }
    }
    
    // MARK: - Navigation

    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let addEditEventVC = segue.destination as? AddEditEventViewController {
            addEditEventVC.event = event
        }
        if let locationVC = segue.destination as? LocationViewController {
            locationVC.eventUser = eventUser
        }
    }
    

}
