//
//  EventsListViewController.swift
//  Teleasystent
//

import UIKit

class EventsListViewController: UIViewController, UITableViewDataSource, UITableViewDelegate, UIPickerViewDataSource, UIPickerViewDelegate, UITextFieldDelegate {
    
    @IBOutlet weak var eventsFilter: UITextField!
    @IBOutlet weak var eventsLabel: BaseUILabel!
    @IBOutlet weak var eventsTableView: UITableView!
    
    var user: AssistantData?
    var refreshListOnAppear = true
    
    static let noAssistantName = Utils.getLocString("addevent_assistant_none")!
    static let noUserName = Utils.getLocString("addevent_user_none")!
    static let noEventTypeName = Utils.getLocString("addevent_type_none")!
    let headerBgColor = UIColor(white: 0.85, alpha: 1)
    let cellBgColor = UIColor(white: 1.0, alpha: 1)
    let cellPressedBgColor = UIColor(white: 0.8, alpha: 1)
    
    private var events = [EventData]()
    private var filteredEvents = [EventData]()
    private var lastEventsUpdateTime: Date?
    private var lastSelectedEvent: EventData?
    private var lastSelectedEventUser: UserData?
    private var sortColumn = SortColumn.Date
    private var selectedFilter = EventsFilter.Future
    private var sortAscending = false
    private var headerCell : EventsListTableViewCell?
    private weak var refreshEventsErrorDialog: UIAlertController?
    private let filterPicker = UIPickerView()
    
    enum EventsFilter: Int, EnumCollection {
        case Future
        case Past
        case Unassigned
    }
    
    enum SortColumn {
        case User
        case Date
        case EventType
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        //eventsTableView.tableFooterView = UIView()
        eventsTableView.dataSource = self
        eventsTableView.delegate = self
        
        filterPicker.tag = 100
        filterPicker.dataSource = self
        filterPicker.delegate = self
        eventsFilter.inputView = filterPicker
        eventsFilter.inputAccessoryView = createPickerToolbar(doneAction: #selector(filterPickerDone))
        eventsFilter.delegate = self
        eventsFilter.tintColor = UIColor.clear
    }
    
    override func viewWillAppear(_ animated: Bool) {
        print("will appear")
        super.viewWillAppear(animated)
        
        navigationItem.leftBarButtonItem = UIBarButtonItem(title: Utils.getLocString("eventslist_signout")!, style: .plain, target: self, action: #selector(onSignOutPressed))
        navigationItem.rightBarButtonItem = UIBarButtonItem(barButtonSystemItem: UIBarButtonSystemItem.refresh, target: self, action: #selector(onRefreshPressed))
        navigationItem.rightBarButtonItem?.accessibilityLabel = Utils.getLocString("eventslist_refresh")
        
        self.title = Utils.getLocString("eventslist_title")
        eventsLabel.text = Utils.getLocString("eventslist_listLabel")
        
        selectedFilter = AppDelegate.getInstance().restoreData.eventsListFilter ?? selectedFilter
        filterPicker.selectRow(selectedFilter.rawValue, inComponent: 0, animated: false)
        pickerView(filterPicker, didSelectRow: selectedFilter.rawValue, inComponent: 0)
        
        UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, eventsFilter)
        
        let _ = AppDelegate.getInstance().serverManager.updateAssistantsIfStale(completion: nil)
        let _ = AppDelegate.getInstance().serverManager.updateEventTypesIfStale(completion: nil)
        if refreshListOnAppear || AppDelegate.getInstance().shouldUpdateDataFromServer(lastUpdate: lastEventsUpdateTime) {
            refreshEvents()
        }
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    func restoreEventIfPossible() {
        if let eventId = AppDelegate.getInstance().restoreData.selectedEventId {
            guard let event = events.first(where: { $0.eventId?.compare(eventId) == .orderedSame }) else {
                return
            }
            onEventSelected(event: event)
        }
    }
    
    func refreshEvents() {
        if let errorDialog = refreshEventsErrorDialog {
            errorDialog.dismiss(animated: false, completion: nil)
            refreshEventsErrorDialog = nil
        }
        
        let loadingVC = LoadingDialogController()
        present(loadingVC, animated: false, completion: nil)
        AppDelegate.getInstance().serverManager.getAllEvents(assistantId: user!.assistantId, userId: nil) { events, error in
            AppDelegate.getInstance().serverManager.getAllEvents(assistantId: AssistantData.assistantNotAssignedId, userId: nil) { naEvents, naError in
                weak var weakSelf = self
                Async.main {
                    loadingVC.dismiss(animated: false, completion: nil)
                    UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, weakSelf?.eventsFilter)
                    if let error = error ?? naError {
                        weakSelf?.handleGetEventsError(error: error)
                    } else if let events = events, let naEvents = naEvents {
                        let allEvents = events + naEvents
                        weakSelf?.lastEventsUpdateTime = Date(timeIntervalSinceNow: 0.0)
                        weakSelf?.reloadEventsList(events: allEvents)
                        weakSelf?.restoreEventIfPossible()
                    } else {
                        weakSelf?.handleGetEventsError(error: NSError(domain: "GetEventError", code: 1, userInfo: nil))
                    }
                    weakSelf?.refreshListOnAppear = false
                }
            }
        }
    }
    
    func reloadEventsList(events: [EventData]?) {
        if let events = events {
            self.events = events
            let myAssistantId = user?.assistantId
            switch selectedFilter {
            case .Future: filteredEvents = events.filter({ $0.assistantId == myAssistantId && ($0.dateObj?.timeIntervalSinceNow ?? -1) >= 0 })
            case .Past: filteredEvents = events.filter({ $0.assistantId == myAssistantId && ($0.dateObj?.timeIntervalSinceNow ?? 1) <= 0 })
            case .Unassigned: filteredEvents = events.filter({ $0.assistantId == AssistantData.assistantNotAssignedId })
            }
            
            switch sortColumn {
            case .User: filteredEvents.sort(by: {
                let left = EventsListViewController.userName(event: $0)
                let right = EventsListViewController.userName(event: $1)
                return sortAscending ? left < right : left > right
            })
            case .Date: filteredEvents.sort(by: { sortAscending ? $0.date < $1.date : $0.date > $1.date })
            case .EventType: filteredEvents.sort(by: {
                let left = EventsListViewController.eventTypeName(event: $0)
                let right = EventsListViewController.eventTypeName(event: $1)
                return sortAscending ? left < right : left > right
            })
            }
        }
        eventsTableView.reloadData()
        updateHeaderUI()
    }
    
    func handleGetEventsError(error: Error) {
        print("GetEvents error: \(error)")
        refreshEventsErrorDialog = presentErrorDialog(error: error) {
            weak var weakSelf = self
            weakSelf?.refreshEvents()
        }
    }
    
    @objc func onSignOutPressed() {
        AppDelegate.getInstance().firebaseAuthManager.logout()
        navigationController?.popViewController(animated: true)
    }
    
    @objc func onRefreshPressed() {
        let _ = AppDelegate.getInstance().serverManager.updateAssistantsIfStale(completion: nil)
        let _ = AppDelegate.getInstance().serverManager.updateEventTypesIfStale(completion: nil)
        refreshEvents()
    }
    
    func createPickerToolbar(doneAction: Selector?) -> UIToolbar {
        let toolBar = UIToolbar()
        toolBar.barStyle = UIBarStyle.default
        toolBar.isTranslucent = true
        toolBar.sizeToFit()
        let flexButton = UIBarButtonItem(barButtonSystemItem: UIBarButtonSystemItem.flexibleSpace, target: nil, action: nil)
        let doneButton = UIBarButtonItem(title: Utils.getLocString("addevent_pickerDone"), style: UIBarButtonItemStyle.done, target: self, action: doneAction)
        doneButton.isAccessibilityElement = true
        doneButton.accessibilityHint = Utils.getLocString("addevent_pickerDoneHint")
        toolBar.setItems([flexButton, doneButton], animated: false)
        toolBar.isUserInteractionEnabled = true
        return toolBar
    }
    
    @objc func filterPickerDone() {
        view.endEditing(true)
        reloadEventsList(events: events)
    }
    
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int {
        return Array(EventsFilter.cases()).count
    }
    
    func numberOfComponents(in pickerView: UIPickerView) -> Int {
        return 1
    }
    
    func pickerView(_ pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String? {
        guard let filter = EventsFilter(rawValue: row) else {
            return ""
        }
        switch filter {
        case .Future: return Utils.getLocString("eventslist_filterFuture")!
        case .Past: return Utils.getLocString("eventslist_filterPast")!
        case .Unassigned: return Utils.getLocString("eventslist_filterUnassigned")!
        }
    }
    
    func pickerView(_ pickerView: UIPickerView, didSelectRow row: Int, inComponent component: Int) {
        let prefix = Utils.getLocString("eventslist_filterPrefix")!
        let suffix = self.pickerView(filterPicker, titleForRow: row, forComponent: component)!
        eventsFilter.text = "\(prefix) - \(suffix)"
        selectedFilter = EventsFilter(rawValue: row)!
        AppDelegate.getInstance().restoreData.eventsListFilter = selectedFilter
        reloadEventsList(events: events)
    }
    
    func textFieldDidBeginEditing(_ textField: UITextField) {
        UIAccessibilityPostNotification(UIAccessibilityLayoutChangedNotification, filterPicker)
    }
    
    func createTableHeader() -> EventsListTableViewCell {
        let cell = eventsTableView.dequeueReusableCell(withIdentifier: "EventsListCell") as! EventsListTableViewCell
        headerCell = cell
        cell.backgroundColor = headerBgColor
        
        let heightMult: CGFloat = 0.4
        cell.user.fitToHeightMult = heightMult
        cell.date.fitToHeightMult = heightMult
        cell.type.fitToHeightMult = heightMult
        
        cell.user.isUserInteractionEnabled = true
        cell.user.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(onUserHeaderClicked)))
        
        cell.date.isUserInteractionEnabled = true
        cell.date.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(onDateHeaderClicked)))
        
        cell.type.isUserInteractionEnabled = true
        cell.type.addGestureRecognizer(UITapGestureRecognizer(target: self, action: #selector(onEventTypeHeaderClicked)))
        
        cell.separatorInset = UIEdgeInsets.zero
        cell.layoutMargins = UIEdgeInsets.zero
        
        updateHeaderUI()
        
        cell.isAccessibilityElement = false
        for label in [cell.user, cell.date, cell.type] as [UILabel] {
            label.isAccessibilityElement = true
            label.accessibilityTraits = UIAccessibilityTraitHeader
            label.accessibilityHint = Utils.getLocString("eventslist_sortHint")
        }
        
        return cell
    }
    
    @objc func onUserHeaderClicked() {
        onHeaderClicked(column: .User)
    }
    
    @objc func onDateHeaderClicked() {
        onHeaderClicked(column: .Date)
    }
    
    @objc func onEventTypeHeaderClicked() {
        onHeaderClicked(column: .EventType)
    }
    
    func onHeaderClicked(column: SortColumn) {
        let oldSortColumn = sortColumn
        sortColumn = column
        if oldSortColumn == sortColumn {
            sortAscending = !sortAscending
        } else {
            sortAscending = true
        }
        reloadEventsList(events: events)
    }
    
    func updateHeaderUI() {
        if let cell = headerCell {
            cell.user.text = Utils.getLocString("eventslist_userLabel")
            cell.date.text = Utils.getLocString("eventslist_dateLabel")
            cell.type.text = Utils.getLocString("eventslist_typeLabel")
            
            for label in [cell.user, cell.date, cell.type] as [UILabel] {
                label.accessibilityLabel = label.text
                label.accessibilityValue = nil
            }
            
            let sortedPrefix = sortAscending ? "\u{25B2} " : "\u{25BC} "
            let sortedValue = Utils.getLocString(sortAscending ? "eventslist_sortAscValue" : "eventslist_sortDescValue")
            switch sortColumn {
            case .User:
                cell.user.text = "\(sortedPrefix)\(cell.user.text ?? "")"
                cell.user.accessibilityValue = sortedValue
            case .Date:
                cell.date.text = "\(sortedPrefix)\(cell.date.text ?? "")"
                cell.date.accessibilityValue = sortedValue
            case .EventType:
                cell.type.text = "\(sortedPrefix)\(cell.type.text ?? "")"
                cell.type.accessibilityValue = sortedValue
            }
        }
    }
    
    
    func numberOfSections(in tableView: UITableView) -> Int {
        return 1
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return filteredEvents.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "EventsListCell", for: indexPath) as! EventsListTableViewCell
        let event = filteredEvents[indexPath.row]
        
        cell.backgroundColor = cellBgColor
        cell.user.text = EventsListViewController.userName(event: event)
        cell.date.text = event.visibleDate
        cell.type.text = EventsListViewController.eventTypeName(event: event)
        
        cell.date.accessibilityLabel = event.visibleDateHint
        
        cell.accessibilityHint = Utils.getLocString("eventslist_selectHint")
        
        cell.separatorInset = UIEdgeInsets.zero
        cell.layoutMargins = UIEdgeInsets.zero
        return cell
    }
    
    static func assistantName(event: EventData) -> String {
        return assistantName(assistantId: event.assistantId ?? AssistantData.assistantNotAssignedId)
    }
    
    static func assistantName(assistantId: String) -> String {
        var name = noAssistantName
        if assistantId != AssistantData.assistantNotAssignedId {
            name = AppDelegate.getInstance().serverManager.assistants?[assistantId]?.name ?? name
        }
        return name
    }
    
    static func userName(event: EventData) -> String {
        return event.userName ?? noUserName
    }
    
    static func eventTypeName(event: EventData) -> String {
        return AppDelegate.getInstance().serverManager.eventTypes?[event.typeId]?.locName ?? noEventTypeName
    }
    
    func tableView(_ tableView: UITableView, viewForHeaderInSection section: Int) -> UIView? {
        return createTableHeader()
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return UIScreen.main.bounds.height / 9
    }
    
    func tableView(_ tableView: UITableView, heightForHeaderInSection section: Int) -> CGFloat {
        return UIScreen.main.bounds.height / 15
    }
    
    func tableView(_ tableView: UITableView, didHighlightRowAt indexPath: IndexPath) {
        let cell = tableView.cellForRow(at: indexPath)
        cell?.backgroundColor = cellPressedBgColor
    }
    
    func tableView(_ tableView: UITableView, didUnhighlightRowAt indexPath: IndexPath) {
        let cell = tableView.cellForRow(at: indexPath)
        cell?.backgroundColor = cellBgColor
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        onEventSelected(event: filteredEvents[indexPath.row])
    }
    
    func onEventSelected(event: EventData) {
        let loadingVC = LoadingDialogController()
        present(loadingVC, animated: false, completion: nil)
        AppDelegate.getInstance().serverManager.getUser(id: event.userId) { userData, error in
            Async.main {
                weak var weakSelf = self
                loadingVC.dismiss(animated: false, completion: nil)
                if let userData = userData {
                    weakSelf?.lastSelectedEvent = event
                    weakSelf?.lastSelectedEventUser = userData
                    weakSelf?.performSegue(withIdentifier: "showEventSegue", sender: self)
                } else {
                    weakSelf?.showGetUserError(error: error ?? NSError(domain: "GetUser", code: 1, userInfo: nil))
                }
            }
        }
    }
    
    func showGetUserError(error: Error) {
        let _ = presentErrorDialog(error: error, okButtonHandler: nil)
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let eventVC = segue.destination as? EventViewController {
            eventVC.event = lastSelectedEvent
            eventVC.eventUser = lastSelectedEventUser
            AppDelegate.getInstance().restoreData.selectedEventId = lastSelectedEvent?.eventId
        }
    }



}
