//
//  LocationViewController.swift
//  Teleprzewodnik
//

import UIKit
import GoogleMaps

class LocationViewController: UIViewController {
    @IBOutlet weak var timeLabel: BaseUILabel!
    @IBOutlet weak var centerButton: UIBaseButton!
    @IBOutlet weak var refreshButton: UIBaseButton!
    @IBOutlet weak var mapView: GMSMapView!
    
    var eventUser: UserData?

    private var lastLocation: LocationData?
    private var marker = GMSMarker()
    private var updateInProgress = false
    private var refreshTimer: Timer?
    private var notificationTimer: Timer?

    override func viewDidLoad() {
        super.viewDidLoad()
        
        let defLatitude = 52.0
        let defLongitude = 19.0
        let zoom: Float = 5.0
        let camera = GMSCameraPosition.camera(withLatitude: defLatitude, longitude: defLongitude, zoom: zoom)
        mapView.camera = camera
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        guard let user = eventUser else {
            navigationController?.popViewController(animated: false)
            return
        }
        
        self.title = Utils.getLocString("location_title")
        
        marker.title = user.name
        
        centerButton.setTitle(Utils.getLocString("location_centerLabel"), for: .normal)
        refreshButton.setTitle(Utils.getLocString("location_refreshLabel"), for: .normal)
        
        updateLocation(newLocation: nil)
        
        if refreshTimer == nil {
            refreshTimer = Timer.scheduledTimer(timeInterval: 5.0, target: self, selector: #selector(onRefreshTimerDidFire(timer:)), userInfo: nil, repeats: true)
            refreshLocation()
        }
        
        if notificationTimer == nil {
            notificationTimer = Timer.scheduledTimer(timeInterval: 15.0, target: self, selector: #selector(onNotificationTimerDidFire(timer:)), userInfo: nil, repeats: true)
            sendLocationNotification()
        }
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        
        if isBeingDismissed || isMovingFromParentViewController {
            AppDelegate.getInstance().restoreData.mapLaunched = false
            if let refreshTimer = refreshTimer {
                refreshTimer.invalidate()
                self.refreshTimer = nil
            }
            if let notificationTimer = notificationTimer {
                notificationTimer.invalidate()
                self.notificationTimer = nil
            }
        }
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    @objc internal func onRefreshTimerDidFire(timer: Timer) {
        refreshLocation()
    }
    
    @IBAction func onCenterClicked(_ sender: UIBaseButton) {
        centerLocation()
    }
    
    private func centerLocation() {
        if let latitude = lastLocation?.curPosLatitude, let longitude = lastLocation?.curPosLongitude {
            mapView.animate(toLocation: CLLocationCoordinate2D(latitude: latitude, longitude: longitude))
        }
    }
    
    @IBAction func onRefreshClicked(_ sender: UIBaseButton) {
        sendLocationNotification()
        refreshLocation()
    }
    
    private func refreshLocation() {
        if !updateInProgress {
            updateInProgress = true
            AppDelegate.getInstance().serverManager.getLocation(userId: eventUser!.userId) { location, error in
                Async.main {
                    weak var weakSelf = self
                    if let error = error {
                        weakSelf?.handleGetLocationError(error: error)
                    } else if let location = location {
                        if let locationError = location.curPosError {
                            weakSelf?.handleLocationError(error: locationError)
                        } else {
                            weakSelf?.updateLocation(newLocation: location)
                        }
                    }
                    weakSelf?.updateInProgress = false
                }
            }
        }
    }
    
    private func updateLocation(newLocation: LocationData?) {
        var timestampStr = "-"
        if let timestampDate = newLocation?.curPosTimestampDate?.toLocalTime() {
            timestampStr = Utils.createDateFormatter(dateStyle: .medium, timeStyle: .medium).string(from: timestampDate)
        }
        timeLabel.text = "\(Utils.getLocString("location_timePrefix")!) \(timestampStr)"
        
        if let latitude = newLocation?.curPosLatitude, let longitude = newLocation?.curPosLongitude {
            marker.position = CLLocationCoordinate2D(latitude: latitude, longitude: longitude)
            marker.map = mapView
            if lastLocation == nil {
                mapView.animate(to: GMSCameraPosition.camera(withTarget: marker.position, zoom: 15.0))
            }
        } else {
            marker.map = nil
        }
        
        lastLocation = newLocation
    }
    
    private func handleGetLocationError(error: Error) {
        timeLabel.text = Utils.isNetworkError(error: error) ? Utils.getLocString("common_networkError") : Utils.getLocString("common_unknownError")
    }
    
    private func handleLocationError(error: Error) {
        var message = Utils.getLocString("common_unknownError")
        if error is LocationDataEmptyError {
            message = Utils.getLocString("location_emptyError")
        } else if error is LocationDataPermissionError {
            message = Utils.getLocString("location_permissionError")
        } else if error is LocationDataLocationServiceError {
            message = Utils.getLocString("location_locationServiceError")
        }
        timeLabel.text = message
    }
    
    public static func sendLocationNotification(userId: String) {
        AppDelegate.getInstance().serverManager.sendLocationNotification(userId: userId) { success, error in
            print("location notification send: \(success)")
        }
    }
    
    private func sendLocationNotification() {
        LocationViewController.sendLocationNotification(userId: eventUser!.userId)
    }
    
    @objc internal func onNotificationTimerDidFire(timer: Timer) {
        sendLocationNotification()
    }
    
}
