//
//  ServerData.swift
//  Teleasystent
//

import Foundation

struct AssistantData: Decodable {
    static let assistantNotAssignedId = "0"
    static let assistantIdParamName = "assistantId"
    
    let assistantId: String
    let name: String
    let skypeName: String
    let mail: String
    
    func getParams() -> [String:String?] {
        return [
            AssistantData.assistantIdParamName: assistantId,
            "name": name,
            "skypeName": skypeName,
            "mail": mail
        ]
    }
}

struct UserData: Decodable {
    static let userIdParamName = "userId"
    
    let userId: String
    let name: String
    let skypeName: String
    let mail: String
    let dog: String
    let stick: String
    
    var hasDog: Bool {
        return dog != "0"
    }
    
    var stickInRightHand: Bool {
        return stick != "0"
    }
    
    init(userId: String, name: String, skypeName: String, mail: String, dog: Bool, stick: Bool) {
        self.userId = userId
        self.name = name
        self.skypeName = skypeName
        self.mail = mail
        self.dog = dog ? "1" : "0"
        self.stick = stick ? "1" : "0"
    }
    
    func getParams() -> [String:String?] {
        return [
            UserData.userIdParamName: userId,
            "name": name,
            "skypeName": skypeName,
            "mail": mail,
            "dog": dog,
            "stick": stick
        ]
    }
}

struct EventData: Decodable {
    static let eventIdParamName = "eventId"
    static let userIdParamName = "userId"
    static let assistantIdParamName = "assistantId"
    static let dateParamName = "date"
    static let periodParamName = "period"
    static let periodPastParamValue = "past"
    static let periodFutureParamValue = "future"
    
    var eventId: String?
    let userId: String
    var assistantId: String?
    let date: String
    let typeId: String
    var description: String?
    var mark: String?
    var userName: String?
    
    var dateObj: Date? {
        return EventData.dateFromString(dateStr: date)
    }
    
    var visibleDate: String {
        if let dateObj = dateObj {
            return Utils.createDateFormatter().string(from: dateObj)
        }
        return date
    }
    
    var visibleDateHint: String {
        if let dateObj = dateObj {
            return Utils.dateToAccessibilityString(date: dateObj)
        }
        return date
    }
    
    var markInt: Int? {
        return Int(mark ?? "")
    }
    
    init(eventId: String?, userId: String, assistantId: String?, date: Date, typeId: String, description: String?, mark: String?) {
        self.eventId = eventId
        self.userId = userId
        self.assistantId = assistantId
        self.date = EventData.dateToString(date: date)
        self.typeId = typeId
        self.description = description
        self.mark = mark
    }
    
    func getParams(includeEventId: Bool, checkAssistantIdChanged: Bool) -> [String:String?] {
        var params = [
            EventData.userIdParamName: userId,
            EventData.dateParamName: date,
            "typeId": typeId,
            "sprawdzajID": checkAssistantIdChanged ? "1" : "0"
        ]
        if let assistantId = assistantId {
            params[EventData.assistantIdParamName] = assistantId
        }
        if includeEventId, let eventId = eventId {
            params[EventData.eventIdParamName] = eventId
        }
        if let description = description {
            params["description"] = description
        }
        if let mark = mark {
            params["mark"] = mark
        }
        return params
    }
    
    static func dateToString(date: Date) -> String {
        return getDateFormatter().string(from: date)
    }
    
    static func dateFromString(dateStr: String) -> Date? {
        return getDateFormatter().date(from: dateStr)
    }
    
    static func getDateFormatter() -> DateFormatter {
        let dateFormatter = DateFormatter()
        dateFormatter.locale = Locale(identifier: "en_US_POSIX")
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        dateFormatter.dateFormat = "yyyy-MM-dd HH:mm:ss"
        return dateFormatter
    }
}

struct EventTypeData: Decodable {
    let typeId: String
    let name: String
    let namePL: String

    var locName: String {
        return Locale.current.languageCode?.lowercased() == "pl" ? namePL : name
    }
    
}

struct LocationData: Decodable {
    let userId: String
    let curPos: String
    let curPosTimestamp: String
    
    var curPosLatitude: Double? {
        let latitudeStr = curPos.split(separator: ",").first
        if let latitudeStr = latitudeStr {
            return Double(latitudeStr.trimmingCharacters(in: .whitespaces))
        }
        return nil
    }
    
    var curPosLongitude: Double? {
        let longitudeStr = curPos.split(separator: ",").last
        if let longitudeStr = longitudeStr {
            return Double(longitudeStr.trimmingCharacters(in: .whitespaces))
        }
        return nil
    }
    
    var curPosError: Error? {
        if curPos.isEmpty {
            return LocationDataEmptyError(domain: "LocationData", code: 1, userInfo: nil)
        } else if curPos.starts(with: "error") {
            if curPos.contains("permission not granted") {
                return LocationDataPermissionError(domain: "LocationData", code: 2, userInfo: nil)
            } else if curPos.contains("location service disabled") {
                return LocationDataLocationServiceError(domain: "LocationData", code: 3, userInfo: nil  )
            }
            return NSError(domain: "LocationData", code: 4, userInfo: nil)
        }
        return nil
    }
    
    var curPosTimestampDate: Date? {
        return EventData.dateFromString(dateStr: curPosTimestamp)
    }
    
    init(userId: String, curPosLatitude: Double, curPosLongitude: Double, curPosTimestamp: Date) {
        self.userId = userId
        self.curPos = "\(curPosLatitude), \(curPosLongitude)"
        self.curPosTimestamp = EventData.dateToString(date: curPosTimestamp)
    }
    
    init(userId: String, error: String, curPosTimestamp: Date) {
        self.userId = userId
        self.curPos = "error: \(error)"
        self.curPosTimestamp = EventData.dateToString(date: curPosTimestamp)
    }
    
    func getParams() -> [String:String?] {
        let params = [
            EventData.userIdParamName: userId,
            "curPos": curPos,
            "curPosTimestamp": curPosTimestamp,
            ]
        return params
    }
}

class LocationDataEmptyError: NSError { }
class LocationDataPermissionError: NSError { }
class LocationDataLocationServiceError: NSError { }

