//
//  ServerManager.swift
//  Teleasystent
//

import Foundation

class ServerManager {
    private let serverAddress: String
    
    var assistants: [String:AssistantData]?
    var eventTypes: [String:EventTypeData]?
    
    private var lastAssistantsUpdateDate: Date?
    private var lastEventTypesUpdateDate: Date?
    
    
    init(serverAddress: String) {
        self.serverAddress = serverAddress
//        updateAssistants(completion: nil)
//        updateEventTypes(completion: nil)
    }
    
    private func doGet(funcName: String, params: [String:String?], completion: ((Data?, URLResponse?, Error?) -> Void)?) {
        var components = URLComponents(string: serverAddress + funcName)!
        components.queryItems = params.map { key,value in URLQueryItem(name: key, value: value) }
//        print(components.url!.absoluteString)
        let task = URLSession.shared.dataTask(with: components.url!) { data, response, error in
            completion?(data, response, error)
        }
        task.resume()
    }
    
    private func doGetWithStringAnswer(funcName: String, params: [String:String?], completion: ((String?, Error?) -> Void)?) {
        doGet(funcName: funcName, params: params) { data, response, error in
            guard let data = data, error == nil else {
                completion?(nil, error ?? NSError(domain: "Unknown error", code: 1, userInfo: nil))
                return
            }
            let answer = String(data: data, encoding: String.Encoding.utf8) ?? ""
            completion?(answer, nil)
        }
    }
    
    private func doGetWithJSONAnswer<T:Decodable>(funcName: String, params: [String:String?], type: T.Type, completion: ((T?, Error?) -> Void)?) {
        doGet(funcName: funcName, params: params) { data, response, error in
            guard let responseData = data, error == nil else {
                completion?(nil, error ?? NSError(domain: "Unknown error", code: 1, userInfo: nil))
                return
            }
            
            var jsonStr = String(data: responseData, encoding: String.Encoding.utf8)!.trimmingCharacters(in: .whitespacesAndNewlines)
            let jsonStrLines = jsonStr.components(separatedBy: .newlines)
            if jsonStrLines.count >= 3 {
                jsonStr = jsonStrLines[jsonStrLines.count - 3]
            }
            jsonStr = jsonStr.htmlDecoded
//            print(jsonStr)
            
            let jsonData = jsonStr.data(using: String.Encoding.utf8)! // responseData
            let result = try? JSONDecoder().decode(T.self, from: jsonData)
            completion?(result, nil)
        }
    }
    
    private func addAssistantHashToParams(params: inout [String:String?]) {
        addHashToParams(params: &params, hashMidParam: params[AssistantData.assistantIdParamName]!)
    }
    
    private func addUserHashToParams(params: inout [String:String?]) {
        addHashToParams(params: &params, hashMidParam: params[UserData.userIdParamName]!)
    }
    
    private func addEventHashToParams(params: inout [String:String?]) {
        addHashToParams(params: &params, hashMidParam: params[EventData.eventIdParamName]!)
    }
    
    private func addHashToParams(params: inout [String:String?], hashMidParam: String?) {
        var randomParam: String?
        if hashMidParam == nil {
            randomParam = String.random(length: 16)
        }
        let hashSource = "657urfj7v2\(hashMidParam ?? randomParam!)75jd25Eg01s"
        let hash = hashSource.utf8.md5.description
        params["h"] = hash
        if let randomParam = randomParam {
            params["r"] = randomParam
        }
    }
    
    func addOrEditAssistant(assistant: AssistantData, completion: ((Bool, Error?) -> Void)?) {
        var params = assistant.getParams()
        addAssistantHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "addOrEditAssistant.php", params: params) { answer, error in
            completion?(answer?.starts(with: "ok") ?? false, error)
        }
    }
    
    func getAssistants(completion: ((Array<AssistantData>?, Error?) -> Void)?) {
        var params = [String:String?]()
        addHashToParams(params: &params, hashMidParam: nil)
        doGetWithJSONAnswer(funcName: "showAssistants.php", params: params, type: Array<AssistantData>.self) { assistants, error in
            completion?(assistants, error)
        }
    }
    
    func addOrEditUser(user: UserData, completion: ((Bool, Error?) -> Void)?) {
        var params = user.getParams()
        addUserHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "addOrEditUser.php", params: params) { answer, error in
            completion?(answer?.starts(with: "ok") ?? false, error)
        }
    }
    
    func getUser(id: String, completion: ((UserData?, Error?) -> Void)?) {
        var params: [String:String?] = [UserData.userIdParamName:id]
        addUserHashToParams(params: &params)
        doGetWithJSONAnswer(funcName: "showUserById.php", params: params, type: Array<UserData>.self) { users, error in
            completion?(users?.first, error)
        }
    }
    
    func getEventTypes(completion: ((Array<EventTypeData>?, Error?) -> Void)?) {
        var params = [String:String?]()
        addHashToParams(params: &params, hashMidParam: nil)
        doGetWithJSONAnswer(funcName: "showTypes.php", params: params, type: Array<EventTypeData>.self) { types, error in
            completion?(types, error)
        }
    }
    
    func addEditEvent(event: EventData, completion: ((EventData, String?, Error?) -> Void)?) {
        var params = event.getParams(includeEventId: true, checkAssistantIdChanged: true)
        addUserHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "addOrEditEvent.php", params: params) { ans, error in
            if let error = error {
                completion?(event, nil, error)
                return
            }
            let okPrefix = "ok:"
            guard let answer = ans, answer.starts(with: okPrefix) else {
                let serverError = (ans?.starts(with: "error:02 proba nadpisania asystenta") ?? false) ? AssistantChangedError(domain: "addEditEvent", code: 1, userInfo: nil) : NSError(domain: "addEditEvent", code: 1, userInfo: nil)
                completion?(event, nil, serverError)
                return
            }
            let eventId = answer.dropFirst(okPrefix.count).trimmingCharacters(in: .whitespacesAndNewlines)
            completion?(event, eventId, nil)
        }
    }
    
    func deleteEvent(id: String, completion: ((Bool, Error?) -> Void)?) {
        var params: [String:String?] = [EventData.eventIdParamName:id]
        addEventHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "deleteEvent.php", params: params) { answer, error in
            completion?(answer?.starts(with: "ok") ?? false, error)
        }
    }
    
    func getPastEvents(assistantId: String?, userId: String?, date: Date?, completion: ((Array<EventData>?, Error?) -> Void)?) {
        getEvents(assistantId: assistantId, userId: userId, date: date, period: EventData.periodPastParamValue, completion: completion)
    }
    
    func getFutureEvents(assistantId: String?, userId: String?, date: Date?, completion: ((Array<EventData>?, Error?) -> Void)?) {
        getEvents(assistantId: assistantId, userId: userId, date: date, period: EventData.periodFutureParamValue, completion: completion)
    }
    
    func getAllEvents(assistantId: String?, userId: String?, completion: ((Array<EventData>?, Error?) -> Void)?) {
        getEvents(assistantId: assistantId, userId: userId, date: nil, period: nil, completion: completion)
        
//        getEvents(assistantId: assistantId, userId: userId, date: nil, period: EventData.periodFutureParamValue) { events, error in
//            guard let futureEvents = events else {
//                completion?(events, error)
//                return
//            }
//            self.getEvents(assistantId: assistantId, userId: userId, date: nil, period: EventData.periodPastParamValue) { pastEvents, pastError in
//                if let pastEvents = pastEvents {
//                    let allEvents = futureEvents + pastEvents.filter({ pEvent in !futureEvents.contains(where: { $0.eventId == pEvent.eventId }) })
//                    completion?(allEvents, nil)
//                } else {
//                    completion?(nil, pastError)
//                }
//            }
//        }
    }
    
    private func getEvents(assistantId: String?, userId: String?, date: Date?, period: String?, completion: ((Array<EventData>?, Error?) -> Void)?) {
        var params = [String:String?]()
        if let period = period {
            params[EventData.periodParamName] = period
        }
        if let assistantId = assistantId {
            params[EventData.assistantIdParamName] = assistantId
        }
        if let userId = userId {
            params[EventData.userIdParamName] = userId
        }
        if let date = date {
            params[EventData.dateParamName] = EventData.dateToString(date: date)
        }
        addHashToParams(params: &params, hashMidParam: nil)
        doGetWithJSONAnswer(funcName: "showEvents.php", params: params, type: Array<EventData>.self) { events, error in
            completion?(events, error)
        }
    }
    
    func getLocation(userId: String, completion: ((LocationData?, Error?) -> Void)?) {
        var params: [String:String?] = [UserData.userIdParamName:userId]
        addUserHashToParams(params: &params)
        doGetWithJSONAnswer(funcName: "showCurPos.php", params: params, type: Array<LocationData>.self) { locations, error in
            completion?(locations?.first, error)
        }
    }
    
    func editLocation(location: LocationData, completion: ((Bool, Error?) -> Void)?) {
        var params = location.getParams()
        addUserHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "editCurPos.php", params: params) { answer, error in
            completion?(answer?.starts(with: "ok") ?? false, error)
        }
    }
    
    func editFCMToken(userId: String, token: String, completion: ((Bool, Error?) -> Void)?) {
        var params: [String:String?] = [UserData.userIdParamName:userId, "fcmToken":token]
        addUserHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "saveTokenaFcm.php", params: params) { answer, error in
            completion?(answer?.starts(with: "ok") ?? false, error)
        }
    }
    
    func sendLocationNotification(userId: String, completion: ((Bool, Error?) -> Void)?) {
        var params: [String:String?] = [UserData.userIdParamName:userId]
        addUserHashToParams(params: &params)
        doGetWithStringAnswer(funcName: "fcm.php", params: params) { answer, error in
            completion?(answer?.starts(with: "ok") ?? false, error)
        }
    }
    
    func updateAssistantsIfStale(completion: ((Array<AssistantData>?, Error?) -> Void)?) -> Bool {
        if AppDelegate.getInstance().shouldUpdateDataFromServer(lastUpdate: lastAssistantsUpdateDate) {
            updateAssistants(completion: completion)
            return true
        }
        return false
    }
    
    func updateAssistants(completion: ((Array<AssistantData>?, Error?) -> Void)?) {
        getAssistants { assistants, error in
            print("updateAssistants done")
            if let assistants = assistants {
                weak var weakSelf = self
                weakSelf?.lastAssistantsUpdateDate = Date(timeIntervalSinceNow: 0.0)
                weakSelf?.assistants = Dictionary(uniqueKeysWithValues: assistants.map { ($0.assistantId, $0) })
            }
            completion?(assistants, error)
        }
    }
    
    func updateEventTypesIfStale(completion: ((Array<EventTypeData>?, Error?) -> Void)?) -> Bool {
        if AppDelegate.getInstance().shouldUpdateDataFromServer(lastUpdate: lastEventTypesUpdateDate) {
            updateEventTypes(completion: completion)
            return true
        }
        return false
    }
    
    func updateEventTypes(completion: ((Array<EventTypeData>?, Error?) -> Void)?) {
        getEventTypes { types, error in
            print("updateEventTypes done")
            if let types = types {
                weak var weakSelf = self
                weakSelf?.lastEventTypesUpdateDate = Date(timeIntervalSinceNow: 0.0)
                weakSelf?.eventTypes = Dictionary(uniqueKeysWithValues: types.map { ($0.typeId, $0) })
            }
            completion?(types, error)
        }
    }
}

class AssistantChangedError: NSError {
    
}
