//
//  StartViewController.swift
//  Teleasystent
//

import UIKit
import FirebaseUI

class StartViewController: UIViewController, FUIAuthDelegate {
    
    var user: AssistantData?
    var proceedInProgress = false
    var assistantsUpdateInProgress = false
    var eventTypesUpdateInProgress = false
    
    private var controllerVisible = false
    private var loginError: Error?
    private var getUserError: Error?
    private weak var authVC: UIViewController?
    private weak var getUserErrorDialog: UIAlertController?
    private var getAssistantsError: Error?

    override func viewDidLoad() {
        super.viewDidLoad()
    }
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        
        self.title = Utils.getLocString("start_title")
        
        NotificationCenter.default.addObserver(self, selector: #selector(self.proceed), name: NSNotification.Name.UIApplicationDidBecomeActive, object: nil)
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        controllerVisible = true
        
        showLoginErrorIfOccurred()
        showGetUserErrorIfOccurred()
        proceed()
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        controllerVisible = false
        super.viewWillDisappear(animated)
        
        NotificationCenter.default.removeObserver(self)
    }
    
    @objc func proceed() {
        if !proceedInProgress {
            if let getUserErrorDialog = getUserErrorDialog {
                getUserErrorDialog.dismiss(animated: false, completion: nil)
                self.getUserErrorDialog = nil
            }
            if let user = AppDelegate.getInstance().firebaseAuthManager.currentUser {
                self.proceedInProgress = true
                let loadingVC = LoadingDialogController()
                present(loadingVC, animated: false, completion: nil)
                
                assistantsUpdateInProgress = true
                AppDelegate.getInstance().serverManager.updateAssistants { assistants, error in
                    weak var weakSelf = self
                    weakSelf!.getAssistantsError = error
                    weakSelf!.assistantsUpdateInProgress = false
                    weakSelf!.proceedGetUser(user: user, loadingVC: loadingVC)
                }
                eventTypesUpdateInProgress = true
                AppDelegate.getInstance().serverManager.updateEventTypes { types, error in
                    weak var weakSelf = self
                    weakSelf!.eventTypesUpdateInProgress = false
                    weakSelf!.proceedGetUser(user: user, loadingVC: loadingVC)
                }
                proceedGetUser(user: user, loadingVC: loadingVC)
            } else {
                authVC = AppDelegate.getInstance().firebaseAuthManager.showLoginVC(srcVC: self)
                showLoginErrorIfOccurred()
            }
        }
    }
    
    func proceedGetUser(user: User, loadingVC: LoadingDialogController) {
        if !assistantsUpdateInProgress && !eventTypesUpdateInProgress {
            let userId = user.uid
            let assistants = AppDelegate.getInstance().serverManager.assistants
            Async.main {
                weak var weakSelf = self
                loadingVC.dismiss(animated: false, completion: nil)
                if weakSelf?.controllerVisible ?? false {
                    if let assistants = assistants {
                        if let assistant = assistants[userId] {
                            weakSelf?.launchEventsListVC(user: assistant)
                        } else {
                            weakSelf?.launchUserInfoVC()
                        }
                    } else {
                        weakSelf?.getUserError = weakSelf?.getAssistantsError ?? NSError(domain: "GetUser", code: 1, userInfo: nil)
                        weakSelf?.showGetUserErrorIfOccurred()
                    }
                }
                weakSelf?.proceedInProgress = false
            }
        }
    }
    
    func authUI(_ authUI: FUIAuth, didSignInWith authDataResult: AuthDataResult?, error: Error?) {
        print("Auth data result: \(authDataResult?.user.description ?? "nil"), error: \(error?.localizedDescription ?? "nil")")
        loginError = error
        showLoginErrorIfOccurred()
    }
    
    func showLoginErrorIfOccurred() {
        if let error = loginError, let authVC = authVC {
            let _ = authVC.presentErrorDialog(error: error) {
                weak var weakSelf = self
                weakSelf?.loginError = nil
            }
        }
    }
    
    func launchTestVC() {
        performSegue(withIdentifier: "showTestSegue", sender: self)
    }
    
    func launchUserInfoVC() {
        performSegue(withIdentifier: "showUserInfoSegue", sender: self)
    }
    
    func launchEventsListVC(user: AssistantData) {
        self.user = user
        performSegue(withIdentifier: "showEventsListSegue", sender: self)
    }
    
    func showGetUserErrorIfOccurred() {
        if let error = getUserError, controllerVisible {
            print("GetUser error: \(error.localizedDescription)")
            Async.main(after: 1.0) {
                weak var weakSelf = self
                weakSelf?.getUserErrorDialog = weakSelf?.presentErrorDialog(error: error) {
                    weakSelf?.getUserError = nil
                    weakSelf?.proceed()
                }
            }
        }
    }

    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    

    
    // MARK: - Navigation
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        if let eventsListVC = segue.destination as? EventsListViewController {
            eventsListVC.user = user
        }
    }
    

}
