//
//  Utils.swift
//  Tattoo
//

import UIKit

class Utils {
    
    static func fitFontSizeToFrameHeight(_ text: UILabel, _ frame: CGRect, heightMultiplier mult:CGFloat = 1) {
        text.font = text.font.withSize(getFitToFrameHeightFontSize(text, frame, heightMultiplier: mult))
    }
    
    static func getFitToFrameHeightFontSize(_ text: UILabel, _ frame: CGRect, heightMultiplier mult:CGFloat = 1) -> CGFloat {
        return frame.height * mult
    }
    
    static func fitFontSizeToFrameWidth(_ text: UILabel, _ frame: CGRect, widthMultiplier mult:CGFloat = 1) {
        if let fontSize = getFitToFrameWidthFontSize(text, frame, widthMultiplier: mult) {
            text.font = text.font.withSize(fontSize)
        }
    }
    
    static func getFitToFrameWidthFontSize(_ text: UILabel, _ frame: CGRect, widthMultiplier mult:CGFloat = 1) -> CGFloat? {
        if let textStr = text.text {
            let textNSStr = NSString(string:textStr)
            let textCurSize = textNSStr.size(withAttributes: [NSAttributedStringKey.font: text.font])
            let fontDstSize = text.font.pointSize * frame.width/textCurSize.width - 0.5
            return fontDstSize * mult
        }
        return nil
    }
    
    static func fitFontSizeToFrame(_ text: UILabel, _ frame: CGRect, heightMultiplier hmult:CGFloat = 1, widthMultiplier wmult:CGFloat = 1) {
        let hsize = getFitToFrameHeightFontSize(text, frame, heightMultiplier: hmult)
        let wsize = getFitToFrameWidthFontSize(text, frame, widthMultiplier: wmult) ?? hsize
        text.font = text.font.withSize(min(hsize, wsize))
    }
    
    static func getLocString(_ key: String, comment: String = "") -> String? {
        var locStr: String? = NSLocalizedString(key, comment: comment)
        if (comment == "" && locStr == key) || (comment.characters.count > 0 && locStr == comment) {
            let baseBundle = Bundle(path: Bundle.main.path(forResource: "Base", ofType: "lproj")!)
            locStr = baseBundle?.localizedString(forKey: key, value: nil, table: nil)
        }
        return locStr
    }
    
    static func mergePaths(_ dir: URL, relPathInDir: String) -> URL {
        return URL(fileURLWithPath: relPathInDir, relativeTo: dir)
    }
    
    static func doAsyncWithLoading(_ controller: UIViewController, task: @escaping ()->(AnyObject?), completion: ((AnyObject?)->())?, loadingText: String = Utils.getLocString("loading_wait") ?? "") {
        //var result: AnyObject?
        let loadingController = LoadingDialogController(message: loadingText)
        controller.present(loadingController, animated: true, completion: {
            Async.background {
                return task()
            }.main { result in
                loadingController.dismiss(animated: true, completion: {
                    completion?(result)
                })
            }
        })
    }
    
    static func doWithLoading(_ controller: UIViewController, task: @escaping ()->(AnyObject?), completion: ((AnyObject?)->())?, loadingText: String = Utils.getLocString("loading_wait") ?? "") {
        let loadingController = LoadingDialogController(message: loadingText)
        controller.present(loadingController, animated: true, completion: {
            let result = task()
            loadingController.dismiss(animated: true, completion: {
                completion?(result)
            })
        })
    }
    
    static func appITunesLink(_ appId: String, useHttp: Bool = false) -> String {
        let prot = useHttp ? "http" : "itms-apps"
        return "\(prot)://itunes.apple.com/app/id\(appId)"
    }
    
    static func openRatePage() -> Bool {
        let appID = AppDelegate.getInstance().appId
        //let urlStr = "itms-apps://itunes.apple.com/app/id\(appID)" // (Option 1) Open App Page
        let urlStr = "itms-apps://itunes.apple.com/app/viewContentsUserReviews?id=\(appID)" // (Option 2) Open App Review Tab
        return UIApplication.shared.openURL(URL(string: urlStr)!)
    }
    
    
    static func shareText(_ controller: UIViewController, text: String) {
//        let items = [text, NSURL(string: "http://bizomobile.com")!, UIImage(named: "effect_edit.png")!]
        shareItems(controller, items: [text as AnyObject])
    }
    
    static func shareItems(_ controller: UIViewController, items: [AnyObject]) {
        let activity = UIActivityViewController(activityItems: items, applicationActivities: nil)
        activity.popoverPresentationController?.sourceView = controller.view.subviews.first
        controller.present(activity, animated: true, completion: nil)
    }
    
    static func createDateFormatter(dateStyle: DateFormatter.Style = .medium, timeStyle: DateFormatter.Style = .short) -> DateFormatter {
        let dateFormatter = DateFormatter()
        dateFormatter.locale = Locale.current
        dateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
        dateFormatter.dateStyle = dateStyle
        dateFormatter.timeStyle = timeStyle
        return dateFormatter
    }
    
    static func dateToAccessibilityString(date: Date) -> String {
        let dateString = createDateFormatter(dateStyle: .long, timeStyle: .none).string(from: date)
        let tempTimeString = createDateFormatter(dateStyle: .none, timeStyle: .short).string(from: date).trimmingCharacters(in: .whitespaces)
        let tempHourString = tempTimeString.prefix(2)
        let hourString = Utils.getLocString("timeAccessibility_hour\(tempHourString)")!
        let tempMinuteString = String(tempTimeString.suffix(2))
        let minuteString = tempMinuteString.starts(with: "0") ? "\(tempMinuteString.prefix(1)) \(tempMinuteString.suffix(1))" : tempMinuteString
        let nightString = tempHourString == "00" ? Utils.getLocString("timeAccessibility_nightSuffix")! : ""
        let timeString = "\(hourString) \(minuteString) \(nightString)".trimmingCharacters(in: .whitespaces)
        return "\(dateString), \(timeString)"
    }
    
    static func isNetworkError(error: Error) -> Bool {
        return error.localizedDescription.lowercased().contains("network") || error.localizedDescription.lowercased().contains(Utils.getLocString("common_network") ?? "network")
    }
    
}

extension NSLayoutConstraint {
    func setMultiplier(_ multiplier:CGFloat) -> NSLayoutConstraint {
        NSLayoutConstraint.deactivate([self])
        
        let newConstraint = NSLayoutConstraint(
            item: firstItem,
            attribute: firstAttribute,
            relatedBy: relation,
            toItem: secondItem,
            attribute: secondAttribute,
            multiplier: multiplier,
            constant: constant)
        
        newConstraint.priority = priority
        newConstraint.shouldBeArchived = self.shouldBeArchived
        newConstraint.identifier = self.identifier
        
        NSLayoutConstraint.activate([newConstraint])
        return newConstraint
    }
}

extension UINavigationController {
    func popViewControllerWithHandler(_ completion: @escaping ()->()) {
        popViewController(animated: true)
        if let coordinator = transitionCoordinator {
            coordinator.animate(alongsideTransition: nil) { _ in
                completion()
            }
        } else {
            completion()
        }
    }
    func pushViewControllerWithHandler(_ viewController: UIViewController, completion: @escaping ()->()) {
        pushViewController(viewController, animated: true)
        if let coordinator = transitionCoordinator {
            coordinator.animate(alongsideTransition: nil) { _ in
                completion()
            }
        } else {
            completion()
        }
    }
}

extension String {
    var htmlDecoded: String {
        let decoded = try? NSAttributedString(data: Data(utf8), options: [
            .documentType: NSAttributedString.DocumentType.html,
            .characterEncoding: String.Encoding.utf8.rawValue
            ], documentAttributes: nil).string
        
        return decoded ?? self
    }
}

extension String {
    var isEmail: Bool {
        let emailRegEx = "[A-Z0-9a-z._%+-]+@[A-Za-z0-9.-]+\\.[A-Za-z]{2,4}"
        let emailTest = NSPredicate(format:"SELF MATCHES %@", emailRegEx)
        return emailTest.evaluate(with: self)
    }
}

extension String {
    static func random(length: Int = 20) -> String {
        let base = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        var randomString: String = ""
        
        for _ in 0..<length {
            let randomValue = arc4random_uniform(UInt32(base.count))
            randomString += "\(base[base.index(base.startIndex, offsetBy: Int(randomValue))])"
        }
        return randomString
    }
}

extension Collection {
    /// Returns the element at the specified index iff it is within bounds, otherwise nil.
    subscript (safe index: Index) -> Element? {
        return indices.contains(index) ? self[index] : nil
    }
}

extension Date {
    func convertToTimeZone(initTimeZone: TimeZone, timeZone: TimeZone) -> Date {
        let delta = TimeInterval(timeZone.secondsFromGMT() - initTimeZone.secondsFromGMT())
        return addingTimeInterval(delta)
    }
    
    // Convert local time to UTC (or GMT)
    func toGlobalTime() -> Date {
        let timezone = TimeZone.current
        let seconds = -TimeInterval(timezone.secondsFromGMT(for: self))
        return Date(timeInterval: seconds, since: self)
    }
    
    // Convert UTC (or GMT) to local time
    func toLocalTime() -> Date {
        let timezone = TimeZone.current
        let seconds = TimeInterval(timezone.secondsFromGMT(for: self))
        return Date(timeInterval: seconds, since: self)
    }
}

extension UIViewController {
    
    func presentErrorDialog(title: String?, message: String?, okButton: String, okButtonHandler: (()->())?) -> UIAlertController {
        let alert = UIAlertController(title: title, message: message, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: okButton, style: .default, handler: { action in
            okButtonHandler?()
        }))
        present(alert, animated: true, completion: nil)
        return alert
    }
    
    func presentErrorDialog(error: Error, okButtonHandler: (()->())?) -> UIAlertController {
        if Utils.isNetworkError(error: error) {
            return presentNetworkErrorDialog(okButtonHandler: okButtonHandler)
        } else if let _ = error as? AssistantChangedError {
            return presentErrorDialog(title: nil, message: Utils.getLocString("common_assistantChangedError"), okButton: Utils.getLocString("common_ok") ?? "OK", okButtonHandler: okButtonHandler)
        } else {
            return presentUnknownErrorDialog(okButtonHandler: okButtonHandler)
        }
    }
    
    func presentNetworkErrorDialog(okButtonHandler: (()->())?) -> UIAlertController {
        return presentErrorDialog(title: nil, message: Utils.getLocString("common_networkError"), okButton: Utils.getLocString("common_ok") ?? "OK", okButtonHandler: okButtonHandler)
    }
    
    func presentUnknownErrorDialog(okButtonHandler: (()->())?) -> UIAlertController {
        return presentErrorDialog(title: nil, message: Utils.getLocString("common_unknownError"), okButton: Utils.getLocString("common_ok") ?? "OK", okButtonHandler: okButtonHandler)
    }
    
    func presentEmptyFieldErrorDialog(fieldName: String, okButtonHandler: (()->())?) -> UIAlertController {
        let msg = "\(Utils.getLocString("common_fieldEmptyError")!) \n \(fieldName)"
        return presentErrorDialog(title: nil, message: msg, okButton: Utils.getLocString("common_ok") ?? "OK", okButtonHandler: okButtonHandler)
    }
    
    func presentWrongFieldErrorDialog(fieldName: String, okButtonHandler: (()->())?) -> UIAlertController {
        let msg = "\(Utils.getLocString("common_fieldWrongError")!) \n \(fieldName)"
        return presentErrorDialog(title: nil, message: msg, okButton: Utils.getLocString("common_ok") ?? "OK", okButtonHandler: okButtonHandler)
    }
}

public protocol EnumCollection : Hashable {}
public extension EnumCollection {
    static func cases() -> AnySequence<Self> {
        typealias S = Self
        return AnySequence { () -> AnyIterator<S> in
            var raw = 0
            return AnyIterator {
                let current : Self = withUnsafePointer(to: &raw) { $0.withMemoryRebound(to: S.self, capacity: 1) { $0.pointee } }
                guard current.hashValue == raw else { return nil }
                raw += 1
                return current
            }
        }
    }
}

